# lingo - payment and billing system
# Copyright (C) 2022  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

import pytz
from django.db.models import Q
from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404
from django.template.defaultfilters import floatformat
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop as N_
from rest_framework import permissions
from rest_framework.views import APIView
from weasyprint import HTML

from lingo.agendas.models import Agenda
from lingo.api import serializers
from lingo.api.utils import APIErrorBadRequest, Response
from lingo.invoicing.models import (
    PAYMENT_INFO,
    DraftInvoice,
    DraftInvoiceLine,
    InjectedLine,
    Invoice,
    PayerError,
    Payment,
    PaymentType,
    Regie,
)


class AgendaCheckTypeList(APIView):
    permission_classes = ()

    def get(self, request, agenda_identifier=None, format=None):
        agenda = get_object_or_404(Agenda, slug=agenda_identifier)

        check_types = []
        if agenda.check_type_group:
            check_types = [
                {'id': x.slug, 'text': x.label, 'kind': x.kind}
                for x in agenda.check_type_group.check_types.filter(disabled=False)
            ]
        return Response({'data': check_types})


agenda_check_type_list = AgendaCheckTypeList.as_view()


class PricingCompute(APIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = serializers.PricingComputeSerializer

    def get(self, request, format=None):
        serializer = self.serializer_class(data=request.query_params)
        if not serializer.is_valid():
            raise APIErrorBadRequest(N_('invalid payload'), errors=serializer.errors)

        return Response({'data': serializer.compute(self.request)})


pricing_compute = PricingCompute.as_view()


class InvoicingRegies(APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, format=None):
        return Response(
            {
                'data': [
                    {'id': regie.slug, 'text': regie.label, 'slug': regie.slug}
                    for regie in Regie.objects.order_by('label')
                ]
            }
        )


invoicing_regies = InvoicingRegies.as_view()


class InvoicingPaymentTypes(APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, regie_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        return Response(
            {
                'data': [
                    {'id': payment_type.slug, 'text': payment_type.label, 'slug': payment_type.slug}
                    for payment_type in PaymentType.objects.filter(regie=regie, disabled=False)
                ]
            }
        )


invoicing_payment_types = InvoicingPaymentTypes.as_view()


class PayerMixin:
    def get_payer_external_id(self, request, regie, nameid=None, payer_external_id=None):
        if payer_external_id:
            return payer_external_id
        if not nameid:
            raise Http404
        try:
            return regie.get_payer_external_id_from_nameid(request, nameid)
        except PayerError:
            raise Http404


class InvoicingInvoices(PayerMixin, APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get_invoices_queryset(self, request, regie):
        payer_external_id = self.get_payer_external_id(
            request=request,
            regie=regie,
            nameid=request.GET.get('NameID'),
            payer_external_id=request.GET.get('payer_external_id'),
        )
        return Invoice.objects.filter(
            Q(pool__isnull=True) | Q(pool__campaign__finalized=True),
            regie=regie,
            remaining_amount__gt=0,
            date_publication__lte=datetime.date.today(),
            payer_external_id=payer_external_id,
        ).order_by('-created_at')

    def get(self, request, regie_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        invoices = self.get_invoices_queryset(request, regie)
        data = []
        for invoice in invoices:
            invoice_data = invoice.normalize(
                for_backoffice=bool(request.GET.get('payer_external_id')),
                label_plus=bool(request.GET.get('include_lines')),
            )
            data.append(invoice_data)
            if bool(request.GET.get('include_lines')) and not invoice_data['disabled']:
                lines = invoice.get_grouped_and_ordered_lines()
                for line in lines:
                    amount = _('%(amount)s€') % {'amount': floatformat(line.remaining_amount, 2)}
                    parts = ['----']
                    if line.agenda:
                        parts.append('%s -' % line.agenda.label)
                    if not line.details:
                        parts.append('%s -' % line.event_date.strftime('%d/%m/%Y'))
                    parts.append(line.label)
                    if line.details.get('check_type_label'):
                        parts.append('- %s' % line.details['check_type_label'])
                    elif line.details.get('status') == 'absence':
                        parts.append('- %s' % _('Absence'))
                    parts.append(_('(amount to pay: %s)') % amount)
                    data.append(
                        {
                            'id': 'line:%s' % line.uuid,
                            'label': ' '.join(parts),
                            'disabled': not bool(line.remaining_amount),
                            'is_line': True,
                            'agenda': line.agenda.label if line.agenda else None,
                            'event_date': line.event_date,
                            'details': line.details,
                            'line_label': line.label,
                            'remaining_amount': line.remaining_amount,
                        }
                    )
        return Response({'data': data})


invoicing_invoices = InvoicingInvoices.as_view()


class InvoicingHistoryInvoices(InvoicingInvoices):
    def get_invoices_queryset(self, request, regie):
        payer_external_id = self.get_payer_external_id(
            request=request,
            regie=regie,
            nameid=request.GET.get('NameID'),
            payer_external_id=request.GET.get('payer_external_id'),
        )
        return Invoice.objects.filter(
            Q(pool__isnull=True) | Q(pool__campaign__finalized=True),
            regie=regie,
            remaining_amount=0,
            date_publication__lte=datetime.date.today(),
            payer_external_id=payer_external_id,
        ).order_by('-created_at')


invoicing_history_invoices = InvoicingHistoryInvoices.as_view()


class InvoicingInvoice(PayerMixin, APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, regie_identifier, invoice_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        payer_external_id = self.get_payer_external_id(
            request=request,
            regie=regie,
            nameid=request.GET.get('NameID'),
            payer_external_id=request.GET.get('payer_external_id'),
        )
        invoice = get_object_or_404(
            Invoice,
            Q(pool__isnull=True) | Q(pool__campaign__finalized=True),
            uuid=invoice_identifier,
            regie=regie,
            date_publication__lte=datetime.date.today(),
            payer_external_id=payer_external_id,
        )
        return Response(
            {'data': invoice.normalize(for_backoffice=bool(request.GET.get('payer_external_id')))}
        )


invoicing_invoice = InvoicingInvoice.as_view()


class InvoicingInvoicePDF(PayerMixin, APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, regie_identifier, invoice_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        payer_external_id = self.get_payer_external_id(
            request=request,
            regie=regie,
            nameid=request.GET.get('NameID'),
            payer_external_id=request.GET.get('payer_external_id'),
        )
        invoice = get_object_or_404(
            Invoice,
            Q(pool__isnull=True) | Q(pool__campaign__finalized=True),
            uuid=invoice_identifier,
            regie=regie,
            date_publication__lte=datetime.date.today(),
            payer_external_id=payer_external_id,
        )
        result = invoice.html()
        html = HTML(string=result)
        pdf = html.write_pdf()
        response = HttpResponse(pdf, content_type='application/pdf')
        response['Content-Disposition'] = 'attachment; filename="%s.pdf"' % invoice.formatted_number
        return response


invoicing_invoice_pdf = InvoicingInvoicePDF.as_view()


class InvoicingInvoicePaymentsPDF(PayerMixin, APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, regie_identifier, invoice_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        payer_external_id = self.get_payer_external_id(
            request=request,
            regie=regie,
            nameid=request.GET.get('NameID'),
            payer_external_id=request.GET.get('payer_external_id'),
        )
        invoice = get_object_or_404(
            Invoice,
            Q(pool__isnull=True) | Q(pool__campaign__finalized=True),
            uuid=invoice_identifier,
            regie=regie,
            date_publication__lte=datetime.date.today(),
            payer_external_id=payer_external_id,
            remaining_amount=0,
        )
        result = invoice.payments_html()
        html = HTML(string=result)
        pdf = html.write_pdf()
        response = HttpResponse(pdf, content_type='application/pdf')
        response['Content-Disposition'] = 'attachment; filename="A-%s.pdf"' % invoice.formatted_number
        return response


invoicing_invoice_payments_pdf = InvoicingInvoicePaymentsPDF.as_view()


class InvoicingInvoicePay(APIView):
    # XXX ?
    authentication_classes = []
    permission_classes = ()

    def post(self, request, regie_identifier, invoice_identifier):
        data = request.data
        order_id = data.get('transaction_id')
        order_date = data.get('transaction_date')
        if order_date:
            try:
                order_date = pytz.utc.localize(datetime.datetime.strptime(order_date, '%Y-%m-%dT%H:%M:%S'))
            except ValueError:
                order_date = None
        invoice = get_object_or_404(
            Invoice,
            Q(pool__isnull=True) | Q(pool__campaign__finalized=True),
            uuid=invoice_identifier,
            regie__slug=regie_identifier,
            remaining_amount__gt=0,
        )
        payment_type = get_object_or_404(PaymentType, regie=invoice.regie, slug='online')
        payment = Payment.make_payment(
            regie=invoice.regie,
            invoices=[invoice],
            amount=invoice.remaining_amount,
            payment_type=payment_type,
            order_id=order_id,
            order_date=order_date,
        )
        return Response(
            {
                'data': {
                    'id': payment.uuid,
                }
            }
        )


invoicing_invoice_pay = InvoicingInvoicePay.as_view()


class InvoicingDraftInvoices(APIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = serializers.DraftInvoiceSerializer

    def post(self, request, regie_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        serializer = self.serializer_class(data=request.data)
        if not serializer.is_valid():
            raise APIErrorBadRequest(N_('invalid payload'), errors=serializer.errors)

        invoice = DraftInvoice.objects.create(regie=regie, **serializer.validated_data)

        return Response({'data': {'draft_invoice_id': str(invoice.uuid)}})


invoicing_draft_invoices = InvoicingDraftInvoices.as_view()


class InvoicingDraftInvoiceLines(APIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = serializers.DraftInvoiceLineSerializer

    def post(self, request, regie_identifier, invoice_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        invoice = get_object_or_404(DraftInvoice, regie=regie, uuid=invoice_identifier, pool__isnull=True)
        serializer = self.serializer_class(data=request.data)
        if not serializer.is_valid():
            raise APIErrorBadRequest(N_('invalid payload'), errors=serializer.errors)

        line = DraftInvoiceLine.objects.create(
            invoice=invoice,
            payer_external_id=invoice.payer_external_id,
            payer_first_name=invoice.payer_first_name,
            payer_last_name=invoice.payer_last_name,
            payer_address=invoice.payer_address,
            payer_demat=invoice.payer_demat,
            payer_direct_debit=invoice.payer_direct_debit,
            **serializer.validated_data,
        )

        return Response({'data': {'draft_line_id': line.pk}})


invoicing_draft_invoice_lines = InvoicingDraftInvoiceLines.as_view()


class InvoicingDraftInvoiceClose(APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def post(self, request, regie_identifier, invoice_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        invoice = get_object_or_404(DraftInvoice, regie=regie, uuid=invoice_identifier, pool__isnull=True)
        final_invoice = invoice.promote()

        return Response({'data': {'invoice_id': str(final_invoice.uuid)}})


invoicing_draft_invoice_close = InvoicingDraftInvoiceClose.as_view()


class InjectedLines(APIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = serializers.InjectedLineSerializer

    def post(self, request, regie_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        serializer = self.serializer_class(data=request.data)
        if not serializer.is_valid():
            raise APIErrorBadRequest(N_('invalid payload'), errors=serializer.errors)

        instance = InjectedLine.objects.create(regie=regie, **serializer.validated_data)
        return Response({'err': 0, 'id': instance.pk})


injected_lines = InjectedLines.as_view()


class InvoicingPayments(PayerMixin, APIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = serializers.PaymentSerializer

    def get(self, request, regie_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        payer_external_id = self.get_payer_external_id(
            request=request,
            regie=regie,
            nameid=request.GET.get('NameID'),
            payer_external_id=request.GET.get('payer_external_id'),
        )
        payments = Payment.objects.filter(
            regie=regie,
            payer_external_id=payer_external_id,
        ).order_by('-created_at')
        return Response(
            {
                'data': [
                    {
                        'id': str(payment.uuid),
                        'display_id': payment.formatted_number,
                        'payment_type': payment.payment_type.label,
                        'amount': payment.amount,
                        'created': payment.created_at.date(),
                        'has_pdf': True,
                    }
                    for payment in payments
                ]
            }
        )

    def post(self, request, regie_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        serializer = self.serializer_class(data=request.data, regie=regie)
        if not serializer.is_valid():
            raise APIErrorBadRequest(N_('invalid payload'), errors=serializer.errors)

        instance = Payment.make_payment(
            regie=regie,
            invoices=serializer._invoices,
            lines=serializer._lines,
            amount=serializer.validated_data['amount'],
            payment_type=serializer.validated_data['payment_type'],
            payment_info={
                k: serializer.validated_data[k] for k, l in PAYMENT_INFO if serializer.validated_data.get(k)
            },
        )
        return Response({'err': 0, 'id': instance.pk})


invoicing_payments = InvoicingPayments.as_view()


class InvoicingPaymentPDF(PayerMixin, APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, regie_identifier, payment_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        payer_external_id = self.get_payer_external_id(
            request=request,
            regie=regie,
            nameid=request.GET.get('NameID'),
            payer_external_id=request.GET.get('payer_external_id'),
        )
        payment = get_object_or_404(
            Payment,
            uuid=payment_identifier,
            regie=regie,
            payer_external_id=payer_external_id,
        )
        result = payment.html()
        html = HTML(string=result)
        pdf = html.write_pdf()
        response = HttpResponse(pdf, content_type='application/pdf')
        response['Content-Disposition'] = 'attachment; filename="%s.pdf"' % payment.formatted_number
        return response


invoicing_payment_pdf = InvoicingPaymentPDF.as_view()
